import logging
import os
import shutil
import subprocess
import tempfile
from io import StringIO
from itertools import chain
from pathlib import Path
from textwrap import dedent
from typing import Callable, List, Optional

import cwl_utils.parser_v1_1 as parser
import pytest
from cwltool.load_tool import fetch_document, resolve_and_validate_document
from dataclasses import dataclass, field
from WDL import Error, parse_document

from aclimatise import Command, Flag, WrapperGenerator, cli_types
from aclimatise.model import CliArgument
from aclimatise.name_generation import NameGenerationError
from aclimatise.yaml import yaml

logging.getLogger("cwltool").setLevel(30)


def skip_not_installed(executable: str):
    """
    Returns a pytest decorator to skip the test if the given executable is not in the path
    """
    return pytest.mark.skipif(
        not shutil.which(executable), reason="{} is not installed".format(executable)
    )


@dataclass(frozen=True)
class HelpText:
    path: str
    """Path to the test data file"""
    cmd: List[str]
    """command-line used to generate this Command"""
    positional: Optional[int] = None
    """Number of positional args"""
    named: Optional[int] = None
    """Number of named args"""
    subcommands: Optional[int] = None
    """Number of subcommands"""
    outputs: Optional[int] = None
    """Number of outputs"""
    output_assertions: List[Callable[[CliArgument], bool]] = field(default_factory=list)
    """
    A list of assertions that must return True for at least one output
    """
    assertions: List[Callable[[Command], bool]] = field(default_factory=list)
    """
    A list of assertions that must return True if the command passed. Used for misc
    assertions that don't fit into the above fields
    """

    def run_assertions(self, cmd: Command, explore=False):
        """
        Run all assertions against the Command that was parsed
        """
        if explore:
            # If we are exploring, we have distinguished between positionals and
            # subcommands, so can make separate assertions
            if self.positional is not None:
                assert len(cmd.positional) == self.positional
            if self.subcommands is not None:
                assert len(cmd.subcommands) == self.subcommands
        else:
            # If we are not exploring, we haven't distinguished between positionals and
            # subcommands, so we have to combine them
            if self.positional is not None:
                assert len(cmd.positional) == self.positional + self.subcommands

        if self.named is not None:
            assert len(cmd.named) == self.named
        if self.outputs is not None:
            assert len(cmd.outputs) == self.outputs
        if self.output_assertions is not None:
            for assertion in self.output_assertions:
                assert any([assertion(out) for out in cmd.outputs])

        for assertion in self.assertions:
            assert assertion(cmd)

        # Check the converters work
        convert_validate(cmd, explore=explore)


all_tests = [
    pytest.param(
        HelpText(
            path="test_data/typeHLA.txt",
            cmd=["typeHLA.js"],
            positional=1,  # filek
            named=209,  # 208 flags with descriptions, and also "-e"
            subcommands=2,  # shell and d8
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools_pl.txt",
            cmd=["samtools.pl"],
            positional=0,
            named=0,
            subcommands=3,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools_bedcov.txt",
            cmd=["samtools", "bedcov"],
            positional=2,
            named=4,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/gth.txt",
            cmd=["gth"],
            positional=0,
            named=37,
            subcommands=0,
            # Actually this has several outputs, but because the help doesn't like argument it's hard to detect this
            # outputs=1,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_closest.txt",
            cmd=["bedtools", "closest"],
            positional=0,
            named=29,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_spacing.txt",
            cmd=["bedtools", "spacing"],
            positional=0,
            named=5,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools_quickcheck.txt",
            cmd=["samtools", "quickcheck"],
            positional=1,
            named=2,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_subtract.txt",
            cmd=["bedtools", "subtract"],
            positional=0,
            named=20,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_window.txt",
            cmd=["bedtools", "window"],
            positional=0,
            named=15,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools_dict.txt",
            cmd=["samtools", "dict"],
            positional=1,
            named=5,
            subcommands=0,
            # outputs=1,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_index.txt",
            cmd=["bwa", "index"],
            positional=1,
            named=4,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_multiinter.txt",
            cmd=["bedtools", "multiinter"],
            positional=0,
            named=8,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_coverage.txt",
            cmd=["bedtools", "coverage"],
            positional=0,
            named=20,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_mem.txt",
            cmd=["bwa", "mem"],
            positional=3,
            named=36,
            subcommands=0,
            outputs=1,  # -o
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_bwt2sa.txt",
            cmd=["bwa", "bwt2sa"],
            named=1,
            positional=2,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_bwtupdate.txt",
            cmd=["bwa", "bwtupdate"],
            named=0,
            positional=1,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/htseq_count.txt",
            cmd=["htseq-count"],
            named=20,
            positional=2,
            subcommands=0,
            # outputs=1,  # Should actually be 2: -c and -o, but we're currently failing -o
            # output_assertions=[
            #     lambda out: "-c" in out.synonyms,
            #     # lambda out: '-o' in out.synonyms,
            # ],
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools.txt",
            cmd=["bedtools"],
            positional=0,
            named=1,
            subcommands=43,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/pisces.txt",
            cmd=["pisces"],
            named=57,
            positional=0,
            subcommands=0,
            outputs=1,
            output_assertions=[
                lambda out: "--outfolder" in out.synonyms
                and out.get_type() == cli_types.CliDir(output=True)
            ],
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/podchecker.txt",
            cmd=["podchecker"],
            named=2,
            positional=1,
            subcommands=0,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa.txt",
            cmd=["bwa"],
            named=0,
            positional=0,
            subcommands=14,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools.txt",
            cmd=["samtools"],
            positional=0,
            named=0,
            subcommands=29,
            outputs=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_random.txt",
            cmd=["bedtools", "random"],
            positional=0,
            named=4,
            subcommands=0,
            outputs=0,
        ),
    ),
    # These last two are really strange, maybe I'll support them eventually
    pytest.param(
        HelpText(
            path="test_data/dinosaur.txt",
            cmd=["dinosaur"],
            positional=0,
            named=24,
            subcommands=0,
        ),
        marks=pytest.mark.xfail,
    ),
    pytest.param(
        HelpText(
            path="test_data/mauve.txt",
            cmd=["mauveAligner"],
            positional=2,
            named=30,
            subcommands=0,
        ),
        marks=pytest.mark.xfail,
    ),
]

all_tests_lookup = {tuple(param.values[0].cmd): param.values[0] for param in all_tests}

all_ids = [" ".join(case.values[0].cmd) for case in all_tests]


def convert_validate(cmd: Command, lang: str = None, explore=True):
    """
    Converts the command into one or all formats, and then validates that they worked
    """
    if lang:
        conv = WrapperGenerator.choose_converter(lang)()
        with tempfile.TemporaryDirectory() as tempd:
            for path, subcommand in conv.generate_tree(cmd, Path(tempd)):
                content = path.read_text()
                validators[lang](content, subcommand, explore=explore)
    else:
        for lang in ("cwl", "wdl", "yml"):
            convert_validate(cmd, lang, explore=explore)


def validate_cwl(cwl: str, cmd: Command = None, explore: bool = True):
    parsed = yaml.load(cwl)
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        tmpfile = tmpdir / parsed["id"]
        tmpfile.write_text(cwl)

        # cwltool doesn't have an official Python API, so run it as a subprocess
        subprocess.run(["cwltool", "--validate", tmpfile])

        # We don't want this to be an actual field hierarchy
        path = Path(parsed["id"])
        assert len(path.parts) == 1

        if cmd:
            assert len(parsed["inputs"]) == len(cmd.positional) + len(cmd.named)
            # We should have all the official outputs, plus stdout
            assert len(parsed["outputs"]) == len(cmd.outputs) + 1


def validate_wdl(wdl: str, cmd: Command = None, explore=True):
    """
    :param explore: If true, we're in explore mode, and we should ignore subcommands
    """
    try:
        doc = parse_document(wdl)
    except Error.SyntaxError as e:
        print(wdl)
        raise e

    task = doc.tasks[0]

    # task.inputs might be None
    cmd_input_names = {inp.name for inp in task.inputs or []}
    # Verify that every parameter has been documented
    wdl_input_names = {meta for meta in task.parameter_meta.keys()}
    assert cmd_input_names == wdl_input_names

    # wdl_input_descriptions = {meta for meta in doc.tasks[0].parameter_meta.values()}
    # cmd_input_descriptions = {arg.description for arg in [*cmd.subcommands, *cmd.named, *cmd.positional]}
    # assert wdl_input_descriptions == cmd_input_descriptions, wdl_input_descriptions ^ cmd_input_descriptions

    # Check that the generated WDL has the correct parameter meta fields
    if cmd:
        cmd_names = [arg.full_name() for arg in [*cmd.named, *cmd.positional]]
        target = len(cmd.named) + len(cmd.positional)
        # If we're not in explore mode, subcommands will be parsed as inputs
        if not explore:
            target += len(cmd.subcommands)
        assert len(task.parameter_meta) == target, "{} vs {}".format(
            task.parameter_meta.keys(), cmd_names
        )
        # We should have all the official outputs, plus stdout
        assert len(task.outputs) == len(cmd.outputs) + 1


def validate_yml(yml: str, cmd: Command = None, explore=True):
    stream = StringIO(yml)
    yaml.load(stream)


def ensure_conda():
    """
    Ensures we're in a conda env, and that the PATH is correctly set
    """
    if "CONDA_PREFIX" not in os.environ:
        raise Exception("Must be in a conda environment")
    bin = os.path.join(os.environ["CONDA_PREFIX"], "bin")
    os.environ["PATH"] = bin + os.pathsep + os.environ["PATH"]


validators = dict(cwl=validate_cwl, wdl=validate_wdl, yml=validate_yml)



"""
Test the casing (snake_case vs camelCase) used by the converters
"""
import pytest

from aclimatise.converter import WrapperGenerator
from aclimatise.model import EmptyFlagArg, Flag


def test_camel_short(camel_gen):
    flag = Flag(
        synonyms=["-t"], description="number of threads [1]", args=EmptyFlagArg()
    )
    names = camel_gen.choose_variable_names([flag], length=3)
    assert names[0].name == "numberOfThreads"


def test_snake_short(snake_gen):
    flag = Flag(
        synonyms=["-t"], description="number of threads [1]", args=EmptyFlagArg()
    )
    names = snake_gen.choose_variable_names([flag], length=2)
    assert "number" in names[0].name
    assert "threads" in names[0].name


def test_camel_long(camel_gen):
    flag = Flag(
        synonyms=["-g", "--genomepaths", "--genomefolders"],
        description="number of threads [1]",
        args=EmptyFlagArg(),
    )
    names = camel_gen.choose_variable_names([flag], length=2)
    assert names[0].name == "genomeFolders"


def test_snake_long(snake_gen):
    flag = Flag(
        synonyms=["-g", "--genomepaths", "--genomefolders"],
        description="number of threads [1]",
        args=EmptyFlagArg(),
    )
    names = snake_gen.choose_variable_names([flag], length=2)
    assert names[0].name == "genome_folders"



import pytest

from aclimatise.cli_types import (
    CliBoolean,
    CliDir,
    CliFile,
    CliFloat,
    CliInteger,
    CliString,
    CliType,
)
from aclimatise.model import CliArgument, EmptyFlagArg, Flag, SimpleFlagArg, infer_type


@pytest.mark.parametrize(
    "string,typ",
    [
        ("", None),
        ("int", CliInteger()),
        ("size", CliInteger()),
        ("length", CliInteger()),
        ("max", CliInteger()),
        ("min", CliInteger()),
        ("str", CliString()),
        ("float", CliFloat()),
        ("decimal", CliFloat()),
        ("bool", CliBoolean()),
        ("file", CliFile()),
        ("path", CliFile()),
        ("input file", CliFile(output=False)),
        ("output file", CliFile(output=True)),
        ("folder", CliDir()),
        ("directory", CliDir()),
        ("output directory", CliDir(output=True)),
        ("blah 23 blub", CliInteger()),
        ("nonsense 23.42", CliFloat()),
        (".42 gibberish", CliFloat()),
        ("1E-5", CliFloat()),
        ("BOOL Output strand bias files, 'true' or 'false'", CliBoolean()),
        ("file to write out dict file [stdout]", CliFile(output=True)),
        ("Filename to output the counts to instead of stdout.", CliFile(output=True)),
        pytest.param(
            "Write out all SAM alignment records into SAM/BAM files (one per input file needed), annotating each line with its feature assignment (as an optional field with tag 'XF'). See the -p option to use BAM instead of SAM.",
            CliFile(output=True),
            marks=pytest.mark.xfail(
                reason="This description doesn't make it clear that it wants an output file. I'm not sure how this could ever be parsed"
            ),
        ),
    ],
)
def test_type_inference(string, typ):
    inferred_type = infer_type(string)
    assert inferred_type == typ


@pytest.mark.parametrize(
    "flag,typ",
    [
        [
            Flag(
                description="Filename to output the counts to instead of stdout.",
                synonyms=["-c", "--counts_output"],
                args=SimpleFlagArg("OUTPUT_FILENAME"),
            ),
            CliFile(output=True),
        ],
        pytest.param(
            Flag(
                description="redirect output to specified file\ndefault: undefined",
                synonyms=["-o"],
                args=EmptyFlagArg(),
            ),
            CliFile(output=True),
            marks=pytest.mark.xfail(
                reason="Because the help doesn't indicate an argument, we can't know that this is an output file"
            ),
        ),
    ],
)
def test_flag_type_inference(flag: CliArgument, typ: CliType):
    inferred_type = flag.get_type()
    assert inferred_type == typ



import os
import shutil
import tempfile
from unittest.mock import Mock, patch

import pytest

from aclimatise import explore_command
from aclimatise.model import Command, Positional

from .util import (
    HelpText,
    all_ids,
    all_tests,
    convert_validate,
    ensure_conda,
    skip_not_installed,
)


@pytest.mark.parametrize("test", all_tests, ids=all_ids)
def test_explore(test: HelpText):
    """
    A comprehensive end-to-end test that tests the parser and converters, after exploring a given command
    """
    if not shutil.which(test.cmd[0]):
        pytest.skip("{} is not installed".format(test.cmd[0]))

    try:
        ensure_conda()
    except:
        pytest.skip("Not in a conda environment")

    # For speed's sake, only explore to depth 2
    command = explore_command(test.cmd, max_depth=1)

    # Check we parsed correctly
    test.run_assertions(command, explore=True)


@skip_not_installed("dinosaur")
@pytest.mark.timeout(360)
def test_explore_dinosaur():
    """
    Python has an issue with killing process trees, whereby the subprocess runs another subprocess.
    This tests that dinosaur
    :return:
    """
    command = explore_command(["dinosaur"], max_depth=1)


@pytest.mark.skipif(not shutil.which("bwa"), reason="bwa is not installed")
def test_explore_bwa():
    """
    This tests specifically that exploring bwa yields a proper bwa mem
    """
    command = explore_command(["bwa"], max_depth=1)

    # Check that we parsed bwa mem correctly
    mem = [cmd for cmd in command.subcommands if cmd.command[1] == "mem"][0]
    assert len(mem.positional) == 3
    assert len(mem.subcommands) == 0
    assert len(mem.named) >= 30


def test_repeat_positionals():
    """
    Test that, if we have multiple duplicate positionals, only the first is tested
    """
    parent = Command(
        command=[],
        positional=[
            Positional(name="a", description="", position=i) for i in range(10)
        ],
    )
    child = Command(command=[])

    with patch("aclimatise.explore_command", new=lambda *args, **kwargs: child):
        with patch("aclimatise.best_cmd", new=Mock(return_value=parent)) as mocked:
            explore_command([])

            # We should only call best_command once, since there's only one unique positional
            assert mocked.call_count == 1



"""
Contains the objects that represent a "type" of data a flag argument might store
"""
import typing
from enum import Enum

from dataclasses import dataclass


@dataclass(unsafe_hash=True)
class CliType:
    """
    A data type used in the command-line
    """

    @property
    def representable(self) -> set:
        """
        Returns a set of types that this type could alternatively be represented as.
        Adds the class's own type to the _representable set
        """
        return self._representable.union({type(self)})

    # The list of types that this specific type could be representable as
    _representable = set()


@dataclass(unsafe_hash=True)
class CliEnum(CliType):
    """
    One of a list of possible options
    """

    enum: Enum
    """
    The possible options as a Python Enum
    """


@dataclass(unsafe_hash=True)
class CliFloat(CliType):
    """
    Takes a floating-point value
    """

    pass


@dataclass(unsafe_hash=True)
class CliInteger(CliType):
    """
    Takes an integer value
    """

    _representable = {CliFloat}


@dataclass(unsafe_hash=True)
class CliString(CliType):
    """
    Takes a string value
    """

    pass


@dataclass(unsafe_hash=True)
class CliBoolean(CliType):
    """
    Takes a boolean value
    """

    pass


@dataclass(unsafe_hash=True)
class CliFileSystemType(CliType):
    """
    Takes a directory / file path
    """

    output: bool = False
    """
    Indicator if it is input or output
    """


@dataclass(unsafe_hash=True)
class CliDir(CliFileSystemType):
    """
    Takes a directory path
    """

    pass


@dataclass(unsafe_hash=True)
class CliFile(CliFileSystemType):
    """
    Takes a file path
    """

    pass


@dataclass(unsafe_hash=True)
class CliDict(CliType):
    """
    Takes a dictionary value
    """

    key: CliType
    """
    Data type of the keys to this dictionary
    """

    value: CliType
    """
    Data type of the values to this dictionary
    """


@dataclass(unsafe_hash=True)
class CliList(CliType):
    """
    Takes a list value
    """

    value: CliType
    """
    Data type of the values in this list
    """


@dataclass(unsafe_hash=True)
class CliTuple(CliType):
    """
    Takes a list of values with a fixed length, possibly each with different types
    """

    values: typing.List[CliType]
    """
    List of types, in order, held within the tuple
    """

    @property
    def homogenous(self):
        """
        A tuple is homogenous if all types in the tuple are the same, aka the set of all types has length 1
        """
        return len(set([type(x) for x in self.values])) == 1



"""
Tests the generate_name function, which converts a paragraph of text into a variable name
"""
from aclimatise.name_generation import generate_name, preprocess


def test_bwa_mem_t():
    name = next(generate_name(preprocess("number of threads [1]")))
    assert len(name) < 5
    assert "number" in name
    assert "threads" in name


def test_bwa_mem_p():
    name = next(generate_name(preprocess("smart pairing (ignoring in2.fq)")))
    assert len(name) <= 3
    assert "smart" in name
    assert "pairing" in name


def test_bwa_mem_r():
    name = next(
        generate_name(
            preprocess("read group header line such as '@RG\tID:foo\tSM:bar' [null]")
        )
    )
    assert len(name) < 5
    assert "read" in name
    # assert 'header' in name


def test_bwa_mem_i():
    name = next(
        generate_name(
            preprocess(
                "specify the mean, standard deviation (10% of the mean if absent), max (4 sigma from the mean if absent) and min of the insert size distribution. FR orientation only. [inferred]"
            )
        )
    )
    assert len(name) < 5
    assert "specify" in name

    # Ideally this would return "mean" first, but the POS engine thinks that "mean" describes "deviation"
    # assert "mean" in name
    assert "deviation" in name


def test_bedtools_coverage_d():
    name = next(
        generate_name(
            preprocess(
                "Report the depth at each position in each A feature. Positions reported are one based. Each position and depth follow the complete A feature."
            )
        )
    )
    assert len(name) < 5
    assert "report" in name
    assert "depth" in name


def test_bedtools_coverage_s():
    name = next(
        generate_name(
            preprocess(
                "Require same strandedness. That is, only report hits in B that overlap A on the _same_ strand. By default, overlaps are reported without respect to strand"
            )
        )
    )
    assert len(name) < 5
    assert "require" in name
    assert "strandedness" in name


def test_bedtools_coverage_g():
    name = next(
        generate_name(
            preprocess(
                "Provide a genome file to enforce consistent chromosome sort order across input files. Only applies when used with -sorted option."
            )
        )
    )
    assert len(name) < 5
    assert "provide" in name
    assert "file" in name


def test_symbol():
    """
    Check that symbols are correctly removed from the output
    """
    name = next(generate_name(preprocess("/genome@ #file$")))
    assert len(name) < 5
    assert "genome" in name
    assert "file" in name


def test_hyphens():
    name = next(generate_name(preprocess("penalty for 5'- and 3'-end clipping [5,5]")))
    assert len(name) < 5
    assert "penalty" in name

    for word in name:
        assert "-" not in word
        assert "[" not in word
        assert "," not in word



"""
Test all the test data files
"""
import pytest
from pkg_resources import resource_filename

from aclimatise import WrapperGenerator, parse_help

from ..util import HelpText, all_tests, convert_validate, validate_cwl, validate_wdl


@pytest.mark.parametrize("test", all_tests)
def test_all(test: HelpText):
    """
    Tests that generate_names can work on real-life Commands without exceeding reasonable system resources
    """
    with open(resource_filename("test", test.path)) as fp:
        help_text = fp.read()

    cmd = parse_help(test.cmd, help_text)

    WrapperGenerator().choose_variable_names([*cmd.positional, *cmd.named])



import inspect
import tempfile
from io import IOBase, StringIO, TextIOBase
from os import PathLike
from pathlib import Path
from typing import Generator, List

from cwl_utils.parser_v1_1 import (
    CommandInputParameter,
    CommandLineBinding,
    CommandLineTool,
    CommandOutputBinding,
    CommandOutputParameter,
)
from dataclasses import dataclass

from aclimatise import cli_types
from aclimatise.converter import NamedArgument, WrapperGenerator
from aclimatise.model import CliArgument, Command, Flag, Positional
from aclimatise.yaml import yaml


@dataclass
class CwlGenerator(WrapperGenerator):
    case = "snake"

    @classmethod
    def format(cls) -> str:
        return "cwl"

    @staticmethod
    def snake_case(words: list):
        return "_".join([word.lower() for word in words])

    @staticmethod
    def to_cwl_type(typ: cli_types.CliType):
        if isinstance(typ, cli_types.CliFile):
            return "File"
        elif isinstance(typ, cli_types.CliDir):
            return "Directory"
        elif isinstance(typ, cli_types.CliString):
            return "string"
        elif isinstance(typ, cli_types.CliFloat):
            return "double"
        elif isinstance(typ, cli_types.CliInteger):
            return "long"
        elif isinstance(typ, cli_types.CliBoolean):
            return "boolean"
        elif isinstance(typ, cli_types.CliEnum):
            return "string"
        elif isinstance(typ, cli_types.CliList):
            return CwlGenerator.to_cwl_type(typ.value) + "[]"
        elif isinstance(typ, cli_types.CliTuple):
            return [CwlGenerator.to_cwl_type(subtype) for subtype in set(typ.values)]
        else:
            raise Exception(f"Invalid type {typ}!")

    def get_inputs(self, names: List[NamedArgument]) -> List[CommandInputParameter]:
        ret = []
        for arg in names:
            assert arg.name != "", arg
            ret.append(
                CommandInputParameter(
                    id="in_" + arg.name,
                    type=self.to_cwl_type(arg.arg.get_type()),
                    inputBinding=CommandLineBinding(
                        position=arg.arg.position
                        if isinstance(arg.arg, Positional)
                        else None,
                        prefix=arg.arg.longest_synonym
                        if isinstance(arg.arg, Flag)
                        else None,
                    ),
                    doc=arg.arg.description,
                )
            )

        return ret

    def get_outputs(self, names: List[NamedArgument]) -> List[CommandOutputParameter]:
        ret = [
            # We default to always capturing stdout
            CommandOutputParameter(
                id="out_stdout",
                type="stdout",
                doc="Standard output stream",
            )
        ]

        for arg in names:
            typ = arg.arg.get_type()
            if isinstance(typ, cli_types.CliFileSystemType) and typ.output:
                ret.append(
                    CommandOutputParameter(
                        id="out_" + arg.name,
                        type=self.to_cwl_type(typ),
                        doc=arg.arg.description,
                        outputBinding=CommandOutputBinding(
                            glob="$(inputs.in_{})".format(arg.name)
                        ),
                    )
                )
        return ret

    def command_to_tool(self, cmd: Command) -> CommandLineTool:
        """
        Outputs the CWL wrapper to the provided file
        """
        inputs: List[CliArgument] = [*cmd.named] + (
            [] if self.ignore_positionals else [*cmd.positional]
        )
        names = self.choose_variable_names(inputs)

        tool = CommandLineTool(
            id=cmd.as_filename + ".cwl",
            baseCommand=list(cmd.command),
            cwlVersion="v1.1",
            inputs=self.get_inputs(names),
            outputs=self.get_outputs(names),
        )

        return tool

    @property
    def suffix(self) -> str:
        return ".cwl"

    def save_to_string(self, cmd: Command) -> str:
        io = StringIO()
        yaml.dump(self.command_to_tool(cmd).save(), io)
        return io.getvalue()

    def save_to_file(self, cmd: Command, path: Path) -> None:
        map = self.command_to_tool(cmd).save()
        with path.open("w") as fp:
            yaml.dump(map, fp)



"""
This module is concerned with running the actual commands so that we can parse their output
"""
import abc
from typing import List


class Executor(abc.ABC):
    """
    An executor is anything that can take a command such as ["bwa"] or
    ["samtools", "sort"] and return the output
    """

    def __init__(self, timeout: int = 10, raise_on_timout=False):
        """
        :param timeout: Amount of inactivity before the execution will be killed
        :param raise_on_timout: If true, execute will raise a TimeoutError if it
            times out
        """
        # Here we initialise all shared parameters that are used by all executors
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timout

    def handle_timeout(self, e: Exception) -> str:
        """
        Subclasses can call this when a timeout has occurred
        :param e: The timeout exception that caused the timeout
        """
        if self.raise_on_timeout:
            raise TimeoutError()
        else:
            return ""

    @abc.abstractmethod
    def execute(self, command: List[str]) -> str:
        """
        Execute a command defined by a list of arguments, and return the result as a string
        """
        pass



from aclimatise.model import Command


def test_reanalyse(samtools_cmd: Command):
    """
    Test the command.reanalyse() method
    """
    reanalysed = samtools_cmd.reanalyse()
    assert reanalysed.help_text == samtools_cmd.help_text
    assert len(reanalysed.subcommands) == len(samtools_cmd.subcommands)

    re_sort = reanalysed["sort"]
    assert len(re_sort.positional) > 0
    assert len(re_sort.named) > 0



"""
Functions for generating WDL from the CLI data model
"""
import re
from os import PathLike
from pathlib import Path
from typing import Generator, Iterable, List, Set, Tuple, Type

from inflection import camelize
from WDL._grammar import keywords
from wdlgen import (
    ArrayType,
    File,
    Input,
    Output,
    ParameterMeta,
    PrimitiveType,
    Task,
    WdlType,
)

from aclimatise import cli_types, model
from aclimatise.converter import NamedArgument, WrapperGenerator
from aclimatise.model import CliArgument, Command, Flag, Positional
from aclimatise.nlp import wordsegment

#: A regex, borrowed from MiniWDL, that ma
WDL_IDENT = re.compile(r"[a-zA-Z][a-zA-Z0-9_]*")
#: Matches all characters we should remove from a WDL identifier
WDL_STRIP = re.compile(r"(^[^a-zA-Z])|([^a-zA-Z0-9_])")


def escape_wdl_str(text: str):
    """
    Escape literal quotes in a Python string, to become suitable for WDL
    """
    return text.replace('"', '\\"').replace("\n", "\\n")


def flag_to_command_input(
    named_flag: NamedArgument, converter: WrapperGenerator
) -> Task.Command.CommandInput:
    args = dict(name=named_flag.name)

    if isinstance(named_flag.arg, model.Flag):
        args.update(dict(optional=named_flag.arg.optional))
        if isinstance(named_flag.arg.args, model.EmptyFlagArg):
            args.update(dict(true=named_flag.arg.longest_synonym, false=""))
        else:
            args.update(
                dict(
                    prefix=named_flag.arg.longest_synonym,
                )
            )
    elif isinstance(named_flag, model.Positional):
        args.update(dict(optional=False, position=named_flag.position))

    return Task.Command.CommandInput.from_fields(**args)


def lowest_common_type(types: Iterable[cli_types.CliType]):
    type_set = {type(t) for t in types}

    if len(type_set) == 1:
        # If there is only one type, use it
        return type_set.pop()

    if (
        len(type_set) == 2
        and cli_types.CliInteger in type_set
        and cli_types.CliFloat in type_set
    ):
        # If they're all numeric, they can be represented as floats
        return cli_types.CliFloat()

    if {
        cli_types.CliDir,
        cli_types.CliDict,
        cli_types.CliFile,
        cli_types.CliTuple,
        cli_types.CliList,
    } & type_set:
        # These complex types cannot be represented in a simpler way
        raise Exception(
            "There is no common type between {}".format(", ".join(type_set))
        )

    else:
        # Most of the time, strings can be used to represent primitive types
        return cli_types.CliString()


class WdlGenerator(WrapperGenerator):
    @property
    def suffix(self) -> str:
        return ".wdl"

    case = "snake"

    @property
    def reserved(self) -> Set[Tuple[str, ...]]:
        # Steal the keywords list from miniWDL
        return {tuple(wordsegment.segment(key)) for key in keywords["1.0"]}

    @classmethod
    def format(cls) -> str:
        return "wdl"

    @classmethod
    def type_to_wdl(cls, typ: cli_types.CliType, optional: bool = False) -> WdlType:
        if isinstance(typ, cli_types.CliString):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        elif isinstance(typ, cli_types.CliFloat):
            return WdlType(PrimitiveType(PrimitiveType.kFloat), optional=optional)
        elif isinstance(typ, cli_types.CliBoolean):
            return WdlType(PrimitiveType(PrimitiveType.kBoolean), optional=optional)
        elif isinstance(typ, cli_types.CliInteger):
            return WdlType(PrimitiveType(PrimitiveType.kInt), optional=optional)
        elif isinstance(typ, cli_types.CliFile):
            return WdlType(PrimitiveType(PrimitiveType.kFile), optional=optional)
        elif isinstance(typ, cli_types.CliDir):
            return WdlType(PrimitiveType(PrimitiveType.kDirectory), optional=optional)
        elif isinstance(typ, cli_types.CliTuple):
            if typ.homogenous:
                return WdlType(
                    ArrayType(
                        cls.type_to_wdl(typ.values[0]), requires_multiple=not optional
                    )
                )
            else:
                return WdlType(
                    ArrayType(
                        cls.type_to_wdl(lowest_common_type(typ.values)),
                        requires_multiple=not optional,
                    )
                )
        elif isinstance(typ, cli_types.CliList):
            return WdlType(
                ArrayType(cls.type_to_wdl(typ.value), requires_multiple=not optional)
            )
        elif isinstance(typ, cli_types.CliEnum):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        else:
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)

    def make_inputs(self, named: Iterable[NamedArgument]) -> List[Input]:
        return [
            Input(
                data_type=self.type_to_wdl(
                    named_arg.arg.get_type(), optional=named_arg.arg.optional
                ),
                name=named_arg.name,
            )
            for named_arg in named
        ]

    def make_command(self, cmd: Command, inputs: List[NamedArgument]) -> Task.Command:
        return Task.Command(
            command=" ".join([WDL_STRIP.sub("_", tok) for tok in cmd.command]),
            inputs=[
                flag_to_command_input(input, self)
                for input in inputs
                if isinstance(input.arg, Positional)
            ],
            arguments=[
                flag_to_command_input(input, self)
                for input in inputs
                if isinstance(input.arg, Flag)
            ],
        )

    def make_parameter_meta(self, named: Iterable[NamedArgument]) -> ParameterMeta:
        params = {}
        for named_arg in named:
            params[named_arg.name] = escape_wdl_str(named_arg.arg.description)

        return ParameterMeta(**params)

    def make_task_name(self, cmd: Command) -> str:
        return camelize(
            "_".join([WDL_STRIP.sub("", token) for token in cmd.command]).replace(
                "-", "_"
            )
        )

    def make_outputs(self, names: List[NamedArgument]) -> List[Output]:
        ret = [
            # We default to always capturing stdout
            Output(data_type=File, name="out_stdout", expression="stdout()")
        ]
        for arg in names:
            typ = arg.arg.get_type()
            if isinstance(typ, cli_types.CliFileSystemType) and typ.output:
                ret.append(
                    Output(
                        data_type=self.type_to_wdl(typ),
                        name="out_" + arg.name,
                        expression='"${{in_{}}}"'.format(arg.name),
                    )
                )

        return ret

    def save_to_string(self, cmd: Command) -> str:
        inputs: List[CliArgument] = [*cmd.named] + (
            [] if self.ignore_positionals else [*cmd.positional]
        )
        names = self.choose_variable_names(inputs)

        tool = Task(
            name=self.make_task_name(cmd),
            command=self.make_command(cmd, names),
            version="1.0",
            inputs=self.make_inputs(names),
            outputs=self.make_outputs(names),
            parameter_meta=self.make_parameter_meta(names),
        )

        return tool.get_string()



import docker
import pytest

from aclimatise.execution.docker import DockerExecutor


def test_docker(bwamem_help):
    client = docker.from_env()
    container = client.containers.run(
        "biocontainers/bwa:v0.7.17_cv1",
        entrypoint=["sleep", "999999999"],
        detach=True,
    )

    exec = DockerExecutor(container)
    output = exec.execute(["bwa", "mem"])
    assert output == bwamem_help
    container.kill()


@pytest.mark.timeout(360)
def test_docker_kill():
    """
    Test that the DockerExecutor can kill the command if it times out
    """
    client = docker.from_env(timeout=99999)
    container = client.containers.run(
        "ubuntu:latest",
        entrypoint=["sleep", "999999999"],
        detach=True,
    )

    exec = DockerExecutor(container)
    output = exec.execute(["sleep", "999999"])
    container.kill()
    assert output == ""


def test_no_output():
    # Check that it doesn't crash when no output is received

    client = docker.from_env()
    container = client.containers.run(
        "quay.io/biocontainers/gadem:1.3.1--h516909a_2",
        entrypoint=["sleep", "9999999"],
        detach=True,
    )
    exec = DockerExecutor(container)
    output = exec.execute(["gadem"])
    container.kill()
    assert output is not None


@pytest.mark.timeout(360)
def test_infinite_output():
    """
    Test that the DockerExecutor can kill the command if it's constantly producing output
    """
    client = docker.from_env(timeout=99999)
    container = client.containers.run(
        "ubuntu:latest",
        entrypoint=["sleep", "999999999"],
        detach=True,
    )

    exec = DockerExecutor(container)
    output = exec.execute(["yes"])
    container.kill()
    assert output.startswith("y")



import random
import string

import pytest
from pkg_resources import resource_filename

from aclimatise import parse_help

from .util import (
    HelpText,
    all_ids,
    all_tests,
    all_tests_lookup,
    convert_validate,
    validate_cwl,
    validate_wdl,
)


@pytest.mark.parametrize("test", all_tests, ids=all_ids)
def test_all(test: HelpText):
    """
    A comprehensive end-to-end test that tests the parser and converters, using the test data files
    """
    with open(resource_filename("test", test.path)) as fp:
        help_text = fp.read()

    cmd = parse_help(test.cmd, help_text)

    # Check that the help text is included in the command
    assert cmd.help_text == help_text

    test.run_assertions(cmd, explore=False)


@pytest.mark.timeout(20)
def test_long_text():
    """
    This tests the case where the parse function is handed an inordinate amount of text. In this case, we shouldn't
    bother parsing, and just return an empty command
    """
    text = "\n".join(
        [
            "".join(
                random.choices(
                    string.ascii_letters + " ",
                    weights=[1] * len(string.ascii_letters) + [5],
                    k=100,
                )
            )
            for i in range(2000)
        ]
    )
    command = parse_help(["some", "command"], text=text)
    assert len(command.positional) == 0
    assert len(command.named) == 0



import spacy
import wordsegment


def prevent_sentence_boundary_detection(doc):
    for token in doc:
        token.is_sent_start = False
    return doc


try:
    nlp = spacy.load("en")
    no_sentences = spacy.load("en")
    no_sentences.add_pipe(
        prevent_sentence_boundary_detection, name="prevent-sbd", before="parser"
    )
except IOError:
    raise Exception(
        "Spacy model doesn't exist! Install it with `python -m spacy download en`"
    )

# We load the spacy and the wordsegment models here as a kind of singleton pattern, to avoid multiple functions loading
# redundant copies

if len(wordsegment.WORDS) == 0:
    wordsegment.load()


def is_sentence(text: str, threshold: float = 0.8) -> bool:
    """
    Returns a bool that indicates if this text is likely a sentence. This should probably be replaced by a machine
    learning classifier in the future
    :param threshold: If the ratio of non-word tokens over word tokens is higher than this, then return False
    """

    doc = no_sentences(text)
    sentence = list(doc.sents)[0]
    non_word_count = 0
    word_count = 0
    for tok in sentence:
        pos = tok.pos_
        if pos == "SPACE":
            # Ignore whitespace
            continue

        if pos in {"X", "SYM", "PUNCT", "NUM"}:
            non_word_count += 1
        word_count += 1

    result = word_count == 0 or non_word_count / word_count < threshold
    return result



from aclimatise.execution.local import LocalExecutor

from ..util import skip_not_installed


@skip_not_installed("bwa")
def test_local(bwamem_help):
    exec = LocalExecutor()
    output = exec.execute(["bwa", "mem"])
    assert output == bwamem_help
